
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <avr/interrupt.h>
#include <avr/io.h>
#include <avr/pgmspace.h>

#include <app/clock_sync.h>
#include <app/dhcp_client.h>
#include <app/httpd.h>
#include <arch/spi.h>
#include <arch/uart.h>
#include <net/arp.h>
#include <net/ethernet.h>
#include <net/hal.h>
#include <net/icmp.h>
#include <net/ip.h>
#include <net/tcp.h>
#include <net/udp.h>
#include <sd/fat16.h>
#include <sd/sd.h>
#include <sd/sd_raw.h>
#include <sys/clock.h>
#include <sys/timer.h>

/**
 * \mainpage mega-eth - Atmel AVR ethernet board
 *
 * This project is a basis for developing microcontroller applications
 * connected to an ethernet network.
 *
 * \section hardware The hardware
 *
 * Hardware components already integrated on the reference design include:
 * - Atmel ATmega128 RISC microcontroller with standard 10-pin ISP header
 * - 64 kByte of external SRAM
 * - USB \<-\> RS232 interface
 * - SD/MMC socket
 * - Ethernet interface with ENC28J60 (IEEE 802.3, 10Base-T)
 *
 * The hardware design is expandable by connecting additional components
 * to the existing pin header. Several digitial I/Os, A/D inputs as well
 * as the standard SPI and I2C (TWI) serial interfaces are available for
 * user-defined purposes.
 *
 * The curcuit board is designed as a two-layer board of size 100mm x 80mm. Most
 * components use SMD packages.
 *
 * \image html mega-eth_rev10.jpg "The mega-eth board running with a memory card and ethernet connected."
 * \image html mega-eth_rev10_layout.png "The layout of the revision 1.0 circuit board."
 *
 * \section software The software
 * The software is written in pure standard ANSI-C. Sure, it might not be the
 * smallest or the fastest one, but I think it is quite flexible.
 *
 * By providing the software for this project under the GPLv2, great
 * flexibility is achieved. The following modules are already available:
 * - \link net_stack TCP/IP and UDP/IP stack \endlink.
 * - \link app_dhcp_client Client DHCP support \endlink for automatically configuring the network interface.
 * - \link app_httpd Small and flexible web server \endlink.
 * - \link sd Mature and well-tested SD/MMC access with FAT16 support \endlink.
 * - Client for retrieving current \link app_clock_sync date and time over network (TIME protocol) \endlink.
 * - Several internal modules for providing a \link sys_clock clock \endlink, \link sys_timer timers \endlink, \link arch_uart UART access \endlink and more.
 *
 * I implemented a simple command prompt which is accessible via the UART at 115200 Baud. With
 * commands similiar to the Unix shell you can setup the network parameters, browse the SD/MMC
 * memory card and more.
 * - <tt>cat \<file\></tt>\n
 *   Writes a hexdump of <tt>\<file\></tt> to the terminal.
 * - <tt>cd \<directory\></tt>\n
 *   Changes current working directory to <tt>\<directory\></tt>.
 * - <tt>clock</tt>\n
 *   Prints date and time from the internal clock.
 * - <tt>clock_sync_start</tt>\n
 *   Starts periodically fetching current date and time from the internet.
 *   \note In the official firmware my own server is contacted. This is for
 *         illustration purposes only. The service is not guaranteed and should
 *         not be used more often than every ten minutes!
 * - <tt>clock_sync_stop</tt>\n
 *   Aborts clock synchronization.
 * - <tt>disk</tt>\n
 *   Shows card manufacturer, status, filesystem capacity and free storage space.
 * - <tt>dhcp_client_start</tt>\n
 *   Starts the DHCP client and tries to configure the network interface.
 * - <tt>dhcp_client_abort</tt>\n
 *   Aborts the DHCP client.
 * - <tt>ip_config [\<ip_address\> [\<ip_netmask\> [\<ip_gateway\>]]]</tt>\n
 *   Displays or changes the network interface configuration.
 * - <tt>ls</tt>\n
 *   Shows the content of the current directory.
 * - <tt>sync</tt>\n
 *   Ensures all buffered data is written to the card.
 *
 * \section adaptation Adapting the software to your needs
 * The software has been written with reusability in mind. Most modules are hardware independent,
 * and hardware abstraction has been used where hardware access is needed. That said, it should
 * be fairly easily adaptible to different hardware configurations.
 *
 * By changing the <tt>MCU*</tt> variables in the Makefile, you can use other Atmel
 * microcontrollers or different clock speeds. Some modules are configurable, the parameters reside
 * in extra files named for example <tt>tcp_config.h</tt> or <tt>ethernet_config.h</tt>.
 * 
 * \section bugs Bugs or comments?
 * If you have comments or found a bug in the software - there might be some
 * of them - you may contact me per mail at feedback@roland-riegel.de.
 *
 * \section copyright Copyright 2006-2008 by Roland Riegel
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License version 2 as published by
 * the Free Software Foundation (http://www.gnu.org/copyleft/gpl.html).
 * At your option, you can alternatively redistribute and/or modify the following
 * files under the terms of the GNU Lesser General Public License version 2.1
 * as published by the Free Software Foundation (http://www.gnu.org/copyleft/lgpl.html):
 * - fat16.c
 * - fat16.h
 * - fat16_config.h
 * - partition.c
 * - partition.h
 * - partition_config.h
 * - sd_raw.c
 * - sd_raw.h
 * - sd_raw_config.h
 */

static void dhcp_client_event_callback(enum dhcp_client_event event);
static void clock_sync_event_callback(enum clock_sync_event event);

static const char* next_word(const char* str);
static uint16_t strtoi(const char* str);
static void strtoaddr(const char* str, uint8_t* ip);

static void print_ip(const uint8_t* ip);
static uint8_t print_disk_info(const struct fat16_fs_struct* fs);

static const uint8_t net_mac[6] = {'F', 'O', 'O', 'B', 'A', 'R'};
static bool net_link_up = false;

static volatile uint8_t timer_events = 0;

int main()
{
    /* enable interrupts */
    sei();

    /* setup uart */
    uart_init();
    /* setup stdio */
    uart_connect_stdio();

    /* booting */
    printf_P(PSTR("mega-eth booting...\n"));

    /* initialize hardware */
    spi_init();
    spi_high_frequency();
    hal_init(net_mac);

    /* initialize sd card */
    sd_raw_init();

    /* initialize ethernet protocol stack */
    ethernet_init(net_mac);
    arp_init();
    ip_init(0, 0, 0);
    icmp_init();
    tcp_init();
    udp_init();

    /* setup interval timer */
    TCCR2 |= (1 << WGM21) | (0 << WGM20) | (1 << CS22) | (0 << CS21) | (1 << CS20);
    TIMSK |= (1 << OCIE2);
    OCR2 = F_CPU / 1024 / 100;

    /* start clock */
    clock_init();

    /* start http server */
    httpd_init(80);

    /* main loop */
    char command[64];
    char* command_pos = command;
    char c = 0x01;
    struct fat16_dir_struct* dir = 0;
    while(1)
    {
        /* handle sd card plugging */
        if(sd_get_root_dir() && !sd_raw_available())
        {
            fat16_close_dir(dir);
            sd_close();
            dir = 0;

            printf_P(PSTR("[sd] card unplugged\n"));
        }
        else if(!sd_get_root_dir() && sd_raw_available())
        {
            switch(sd_open())
            {
                case SD_ERROR_NONE:
                    dir = sd_get_root_dir();

                    printf_P(PSTR("[sd] card initialized\n"));
                    break;
                case SD_ERROR_INIT:
                    printf_P(PSTR("[sd] MMC/SD initialization failed\n"));
                    break;
                case SD_ERROR_PARTITION:
                    printf_P(PSTR("[sd] opening partition failed\n"));
                    break;
                case SD_ERROR_FS:
                    printf_P(PSTR("[sd] opening filesystem failed\n"));
                    break;
                case SD_ERROR_ROOTDIR:
                    printf_P(PSTR("[sd] opening root directory failed\n"));
                    break;
                default:
                    printf_P(PSTR("[sd] unknown error\n"));
                    break;
            }
        }

        /* handle network link changes */
        if(!net_link_up && hal_link_up())
        {
            dhcp_client_start(dhcp_client_event_callback);
            net_link_up = true;
        }
        else if(net_link_up && !hal_link_up())
        {
            dhcp_client_abort();
            net_link_up = false;

            ip_set_address(0);
            ip_set_netmask(0);
            ip_set_gateway(0);
        }

        /* receive packets */
        while(ethernet_handle_packet());

        cli();
        uint8_t timer_ticks = timer_events;
        timer_events = 0;
        sei();

        /* handle timeouts */
        while(timer_ticks--)
            timer_interval();

        /* provide a simple shell */
        if(c == 0x01)
        {
            /* print prompt */
            uart_putc('>');
            uart_putc(' ');

            /* next character begins new command */
            command_pos = command;
            command[0] = '\0';
        }

        c = uart_getc_try();
        if(c == '\0')
            continue;

        /* handle backspace */
        if(c == 0x08 || c == 0x7f)
        {
            if(command_pos == command)
                continue;

            *--command_pos = '\0';

            uart_putc(0x08);
            uart_putc(' ');
            uart_putc(0x08);

            continue;
        }

        uart_putc(c);
        if(c != '\n')
            *command_pos++ = c;
        if(c == '\n' ||
           command_pos == &command[sizeof(command) - 1]
          )
        {
            *command_pos = '\0';

            /* prepare for reading next command */
            c = 0x01;

            /* execute command */
            if(command[0] == '\0')
            {
                /* do nothing */
            }
            else if(strncmp_P(command, PSTR("cat "), 4) == 0)
            {
                const char* str_file = &command[4];

                /* search file in current directory and open it */
                struct fat16_file_struct* fd = sd_open_file_in_dir(dir, str_file);
                if(!fd)
                {
                    printf_P(PSTR("[sd] error opening %s\n"), str_file);
                    continue;
                }

                /* print file contents */
                uint8_t buffer[8];
                uint32_t offset = 0;
                while(fat16_read_file(fd, buffer, sizeof(buffer)) > 0)
                {
                    printf_P(PSTR("0x%08x:"));
                    for(uint8_t i = 0; i < 8; ++i)
                    {
                        uart_putc(' ');
                        uart_putc_hex(buffer[i]);
                    }
                    uart_putc('\n');
                    offset += 8;
                }

                fat16_close_file(fd);
            }
            else if(strncmp_P(command, PSTR("cd "), 3) == 0)
            {
                const char* str_dir = &command[3];

                /* change directory */
                struct fat16_dir_entry_struct subdir_entry;
                if(sd_find_file_in_dir(dir, str_dir, &subdir_entry))
                {
                    struct fat16_dir_struct* dir_new = fat16_open_dir(sd_get_fs(), &subdir_entry);
                    if(dir_new)
                    {
                        if(dir != sd_get_root_dir())
                            fat16_close_dir(dir);
                        dir = dir_new;
                        continue;
                    }
                }

                printf_P(PSTR("[sd] directory not found: %s\n"), str_dir);
            }
            else if(strcmp_P(command, PSTR("clock")) == 0)
            {
                struct clock_date_time dt;
                clock_interpret(clock_get(), &dt);

                printf_P(PSTR("%02u.%02u.%04u %02u:%02u:%02u GMT\n"),
                         dt.day,
                         dt.month,
                         dt.year,
                         dt.hour,
                         dt.minute,
                         dt.second
                        );
            }
            else if(strcmp_P(command, PSTR("clock_sync_start")) == 0)
            {
                clock_sync_start(clock_sync_event_callback);
            }
            else if(strcmp_P(command, PSTR("clock_sync_abort")) == 0)
            {
                clock_sync_abort();
            }
            else if(strcmp_P(command, PSTR("disk")) == 0)
            {
                if(!print_disk_info(sd_get_fs()))
                    printf_P(PSTR("[sd] error reading disk info\n"));
            }
            else if(strcmp_P(command, PSTR("dhcp_client_start")) == 0)
            {
                /* configure ip via dhcp */

                if(!dhcp_client_start(dhcp_client_event_callback))
                    printf_P(PSTR("[dhcp] client failed\n"));
            }
            else if(strcmp_P(command, PSTR("dhcp_client_abort")) == 0)
            {
                /* stop the dhcp client */

                dhcp_client_abort();
            }
            else if(strcmp_P(command, PSTR("ip_config")) == 0)
            {
                /* display IP configuration */

                printf_P(PSTR("[ip] ip     : "));
                print_ip(ip_get_address());
                printf_P(PSTR("\n[ip] netmask: "));
                print_ip(ip_get_netmask());
                printf_P(PSTR("\n[ip] gateway: "));
                print_ip(ip_get_gateway());
                printf_P(PSTR("\n"));
            }
            else if(strncmp_P(command, PSTR("ip_config "), 10) == 0)
            {
                /* change IP configuration */

                uint8_t ip[4];
                strtoaddr(&command[10], ip);
                ip_set_address(ip);

                const char* word = next_word(&command[10]);
                strtoaddr(word, ip);
                ip_set_netmask(ip);

                word = next_word(word);
                strtoaddr(word, ip);
                ip_set_gateway(ip);
            }
            else if(strcmp_P(command, PSTR("ls")) == 0)
            {
                /* print directory listing */
                struct fat16_dir_entry_struct dir_entry;
                while(fat16_read_dir(dir, &dir_entry))
                {
                    uint8_t spaces = sizeof(dir_entry.long_name) - strlen(dir_entry.long_name) + 4;

                    printf_P(PSTR("%s%c"), dir_entry.long_name, dir_entry.attributes & FAT16_ATTRIB_DIR ? '/' : ' ');
                    while(spaces--)
                        uart_putc(' ');
                    printf_P(PSTR("%lu\n"), dir_entry.file_size);
                }
            }
            else if(strcmp_P(command, PSTR("sync")) == 0)
            {
                if(!sd_raw_sync())
                    printf_P(PSTR("[sd] error syncing disk\n"));
            }
            else
            {
                printf_P(PSTR("unknown command: %s\n"), command);
            }
        }
    }
    
    while(1);
}

ISR(TIMER2_COMP_vect)
{
    /* fires 100 times a second / every 10ms */
    ++timer_events;
}

void dhcp_client_event_callback(enum dhcp_client_event event)
{
    switch(event)
    {
        case DHCP_CLIENT_EVT_LEASE_ACQUIRED:
            printf_P(PSTR("[dhcp] lease acquired\n"));
            break;
        case DHCP_CLIENT_EVT_LEASE_DENIED:
            printf_P(PSTR("[dhcp] lease denied\n"));
            break;
        case DHCP_CLIENT_EVT_LEASE_EXPIRING:
            printf_P(PSTR("[dhcp] lease expiring\n"));
            break;
        case DHCP_CLIENT_EVT_LEASE_EXPIRED:
            printf_P(PSTR("[dhcp] lease expired\n"));
            break;
        case DHCP_CLIENT_EVT_TIMEOUT:
            printf_P(PSTR("[dhcp] timeout\n"));
            break;
        case DHCP_CLIENT_EVT_ERROR:
            printf_P(PSTR("[dhcp] error\n"));
            break;
    }
}

void clock_sync_event_callback(enum clock_sync_event event)
{
    switch(event)
    {
        case CLOCK_SYNC_EVT_ERROR:
            printf_P(PSTR("[clock] sync error\n"));
            break;
        case CLOCK_SYNC_EVT_TIMEOUT:
            printf_P(PSTR("[clock] sync timeout\n"));
            break;
        case CLOCK_SYNC_EVT_SUCCESS:
            printf_P(PSTR("[clock] sync success\n"));
            break;
    }
}

const char* next_word(const char* str)
{
    while(*str != '\0' && *str != ' ')
        ++str;
    if(*str != '\0')
        ++str;

    return str;
}

uint16_t strtoi(const char* str)
{
    uint16_t l = 0;
    while(*str >= '0' && *str <= '9')
        l = l * 10 + (*str++ - '0');

    return l;
}

void strtoaddr(const char* str, uint8_t* ip)
{
    memset(ip, 0, 4);

    for(uint8_t i = 0; i < 4; ++i)
    {
        ip[i] = strtoi(str);

        while(1)
        {
            if(*str >= '0' && *str <= '9')
            {
                ++str;
            }
            else if(*str == '.')
            {
                ++str;
                break;
            }
            else
            {
                return;
            }
        }
    }
}

void print_ip(const uint8_t* ip)
{
    printf_P(PSTR("%i.%i.%i.%i"), ip[0], ip[1], ip[2], ip[3]);
}

uint8_t print_disk_info(const struct fat16_fs_struct* fs)
{
    if(!fs)
        return 0;

    struct sd_raw_info disk_info;
    if(!sd_raw_get_info(&disk_info))
        return 0;

    printf_P(PSTR("manuf:  0x%02x\n"), disk_info.manufacturer);
    printf_P(PSTR("oem:    %s\n"), (char*) disk_info.oem);
    printf_P(PSTR("prod:   %s\n"), (char*) disk_info.product);
    printf_P(PSTR("rev:    %02x\n"), disk_info.revision);
    printf_P(PSTR("serial: 0x%04lx\n"));
    printf_P(PSTR("date:   %02u/%02u\n"), disk_info.manufacturing_month, disk_info.manufacturing_year);
    printf_P(PSTR("size:   %lu\n"), disk_info.capacity);
    printf_P(PSTR("copy:   %u\n"), disk_info.flag_copy);
    printf_P(PSTR("wr.pr.: %u/%u\n"), disk_info.flag_write_protect_temp, disk_info.flag_write_protect);
    printf_P(PSTR("format: %u\n"), disk_info.format);
    printf_P(PSTR("free:   %lu/%lu\n"), fat16_get_fs_free(fs), fat16_get_fs_size(fs));

    return 1;
}

